//////////////////////////////////////////////////////////////////////////////////////////////////
// File Name: Compass.cs
//      Date: 06/01/2006
// Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
// Permission to use, copy, modify, and distribute this Program and its documentation,
//  if any, for any purpose and without fee is hereby granted, provided that:
//   (i) you not charge any fee for the Program, and the Program not be incorporated
//       by you in any software or code for which compensation is expected or received;
//   (ii) the copyright notice listed above appears in all copies;
//   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
//   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
//          pertaining to distribution of the Program without specific, written prior permission. 
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Drawing;

namespace AntGP
{
    public class Compass
    {
        /// <summary>
        /// This static class is used for compass directions, and simple logic for determine adjacent positions.
        /// </summary>
        static Compass()
        {
        }

        /// <summary>
        /// The four compass directions
        /// </summary>
        public enum CompassDirection {
            /// <summary>North direction - top of screen.</summary>
            North,
            /// <summary>North direction - right side of screen.</summary>
            East,
            /// <summary>South direction - bottom of screen.</summary>
            South,
            /// <summary>North direction - left side of screen.</summary>
            West,
            /// <summary>No direction - used as return value in some functions where there was no valid compass direction found to satisfy the request.</summary>
            None
        };

        /// <summary>
        /// Gets the direction to the left of the current direction.
        /// </summary>
        /// <param name="currentDirection">The current direction the ant is facing.</param>
        /// <returns>The direction to the left of the ant.</returns>
        public static CompassDirection LeftDirection(CompassDirection currentDirection)
        {
            if (currentDirection == CompassDirection.North)
            {
                return CompassDirection.West;
            }
            else if (currentDirection == CompassDirection.West)
            {
                return CompassDirection.South;
            }
            else if (currentDirection == CompassDirection.South)
            {
                return CompassDirection.East;
            }
            else
            {
                return CompassDirection.North;
            }
        }

        /// <summary>
        /// Gets the direction to the right of the current direction.
        /// </summary>
        /// <param name="currentDirection">The current direction the ant is facing.</param>
        /// <returns>The direction to the right of the ant.</returns>
        public static CompassDirection RightDirection(CompassDirection currentDirection)
        {
            if (currentDirection == CompassDirection.North)
            {
                return CompassDirection.East;
            }
            else if (currentDirection == CompassDirection.East)
            {
                return CompassDirection.South;
            }
            else if (currentDirection == CompassDirection.South)
            {
                return CompassDirection.West;
            }
            else
            {
                return CompassDirection.North;
            }
        }

        /// <summary>
        /// Gets the direction that is the opposite of the current direction.
        /// </summary>
        /// <param name="currentDirection">The current direction the ant is facing.</param>
        /// <returns>The direction behind the ant.</returns>
        public static CompassDirection ReverseDirection(CompassDirection currentDirection)
        {
            if (currentDirection == CompassDirection.North)
            {
                return CompassDirection.South;
            }
            else if (currentDirection == CompassDirection.East)
            {
                return CompassDirection.West;
            }
            else if (currentDirection == CompassDirection.West)
            {
                return CompassDirection.East;
            }
            else
            {
                return CompassDirection.North;
            }
        }

        /// <summary>
        /// Gets the map position that is one square away from the current location in the provided direction.
        /// </summary>
        /// <param name="currentLocation">The current location of the ant.</param>
        /// <param name="mapSize">The size of the map. Used to determine if the adjacent location would be off the map.</param>
        /// <param name="direction">The direction to get the adjacent location for.</param>
        /// <returns>The point of the adjacent location, or Map.NullPoint if the adjacent location is off the edge of the map.</returns>
        public static Point AdjacentLocation(Point currentLocation, Size mapSize, CompassDirection direction)
        {
            switch (direction)
            {
                case CompassDirection.North:
                    if (currentLocation.Y > 0)
                    {
                        return new Point(currentLocation.X, currentLocation.Y-1);
                    }
                    break;
                case CompassDirection.South:
                    if (currentLocation.Y < mapSize.Height-1)
                    {
                        return new Point(currentLocation.X, currentLocation.Y+1);
                    }
                    break;
                case CompassDirection.West:
                    if (currentLocation.X > 0)
                    {
                        return new Point(currentLocation.X-1, currentLocation.Y);
                    }
                    break;
                case CompassDirection.East:
                    if (currentLocation.X < mapSize.Width-1)
                    {
                        return new Point(currentLocation.X+1, currentLocation.Y);
                    }
                    break;
            }

            return Map.NullPoint;
        }

        /// <summary>
        /// Gets the map position that is one square to the left of the current location.
        /// </summary>
        /// <param name="currentLocation">The current location of the ant.</param>
        /// <param name="mapSize">The size of the map. Used to determine if the adjacent location would be off the map.</param>
        /// <param name="direction">The current direction that the ant is facing.</param>
        /// <returns>The point that is one square left of the ants location, or Map.NullPoint if the left location is off the edge of the map.</returns>
        public static Point LeftLocation(Point currentLocation, Size mapSize, CompassDirection direction)
        {
            return AdjacentLocation(currentLocation, mapSize, LeftDirection(direction));
        }

        /// <summary>
        /// Gets the map position that is one square to the right of the current location.
        /// </summary>
        /// <param name="currentLocation">The current location of the ant.</param>
        /// <param name="mapSize">The size of the map. Used to determine if the adjacent location would be off the map.</param>
        /// <param name="direction">The current direction that the ant is facing.</param>
        /// <returns>The point that is one square right of the ants location, or Map.NullPoint if the right location is off the edge of the map.</returns>
        public static Point RightLocation(Point currentLocation, Size mapSize, CompassDirection direction)
        {
            return AdjacentLocation(currentLocation, mapSize, RightDirection(direction));
        }

        /// <summary>
        /// Gets the map position that is one square behind the current location.
        /// </summary>
        /// <param name="currentLocation">The current location of the ant.</param>
        /// <param name="mapSize">The size of the map. Used to determine if the adjacent location would be off the map.</param>
        /// <param name="direction">The current direction that the ant is facing.</param>
        /// <returns>The point that is one square behind the ants location, or Map.NullPoint if the reverse location is off the edge of the map.</returns>
        public static Point BehindLocation(Point currentLocation, Size mapSize, CompassDirection direction)
        {
            return AdjacentLocation(currentLocation, mapSize, ReverseDirection(direction));
        }

        /// <summary>
        /// Gets a random compass direction.
        /// </summary>
        /// <param name="options">The Genetic Program options contain the Random object.
        /// The Random is passed around so that with a fixed seed the user can get the same exact results between successive runs.</param>
        /// <returns>A random direction.</returns>
        public static Compass.CompassDirection RandomDirection(GeneticProgramOptions options)
        {
            switch (options.InterpreterRandom(4))
            {
                case 0: return Compass.CompassDirection.North;
                case 1: return Compass.CompassDirection.East;
                case 2: return Compass.CompassDirection.South;
                case 3: return Compass.CompassDirection.West;
            }

            return Compass.CompassDirection.North;
        }
    }
}
