//////////////////////////////////////////////////////////////////////////////////////////////////
// File Name: Evolver.cs
//      Date: 06/01/2006
// Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
// Permission to use, copy, modify, and distribute this Program and its documentation,
//  if any, for any purpose and without fee is hereby granted, provided that:
//   (i) you not charge any fee for the Program, and the Program not be incorporated
//       by you in any software or code for which compensation is expected or received;
//   (ii) the copyright notice listed above appears in all copies;
//   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
//   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
//          pertaining to distribution of the Program without specific, written prior permission. 
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Text;

namespace AntGP
{
    /// <summary>
    /// This class performs the actual Genetic Program logic,  including crossover and mutation.
    /// It uses the interpreter to run the programs to test their fitness.
    /// </summary>
    class Evolver
    {
        /// <summary>
        /// Represents the left or right child node of a given Program Node.
        /// </summary>
        enum ChildSide
        {
            /// <summary>The left child of a given node in the binary program tree.</summary>
            Left,
            /// <summary>The right child of a given node in the binary program tree.</summary>
            Right,
            /// <summary>Either no child node has been selected, or there is no child node.</summary>
            None
        }

        /// <summary>
        /// Before performing the actual crossover or mutation, the evolver must select a random node in the program treen.
        /// </summary>
        class InformationForNodeToActOn
        {
            /// <summary>
            /// Private copy of the parent node of the node that the evolver will perform crossover or mutation on.
            /// </summary>
            private ProgramNode parentNode = null;
            /// <summary>
            /// The parent node of the node that the evolver will perform crossover or mutation on.
            /// </summary>
            /// <remarks>To perform crossover on 2 nodes, we actually need to change the child links on the parent nodes.</remarks>
            public ProgramNode ParentNode
            {
                set
                {
                    parentNode = value;
                }
                get
                {
                    return parentNode;
                }
            }

            /// <summary>
            /// Private copy of the node that the evolver will perform crossover or mutation on.
            /// </summary>
            private ProgramNode childNode = null;
            /// <summary>
            /// The node that the evolver will perform crossover or mutation on.
            /// </summary>
            public ProgramNode ChildNode
            {
                set
                {
                    childNode = value;
                }
                get
                {
                    return childNode;
                }
            }

            /// <summary>
            /// Private that indicates which side of the parent binary tree node that the child is on.
            /// </summary>
            private ChildSide side = ChildSide.None;
            /// <summary>
            /// Indicates which side of the parent binary tree node that the child is on.
            /// </summary>
            public ChildSide Side
            {
                set
                {
                    side = value;
                }
                get
                {
                    return side;
                }
            }
        }

        /// <summary>
        /// The set of all programs being evolved.
        /// </summary>
        Population population = null;

        /// <summary>
        /// Number of times that crossover has been performed on the population.
        /// </summary>
        int iteration = 0;

        /// <summary>
        /// Private Genetic Program options that are used to create the initial population, run the evolver, and determine fitness.
        /// </summary>
        private GeneticProgramOptions options;
        /// <summary>
        /// Genetic Program options that are used to create the initial population, run the evolver, and determine fitness.
        /// </summary>
        public GeneticProgramOptions Options
        {
            set
            {
                options = value;
            }
            get
            {
                return options;
            }
        }

        /// <summary>
        /// Is the genetic programming running. Set this to false to stop the evolver thread.
        /// </summary>
        private bool running;
        /// <summary>
        /// Is the genetic programming running. Set this to false to stop the evolver thread.
        /// </summary>
        public bool Running
        {
            set
            {
                running = value;
            }
            get
            {
                return running;
            }
        }

        /// <summary>
        /// The best program found by the Genetic Program so far.
        /// </summary>
        public Program BestProgram
        {
            get
            {
                if (population != null)
                {
                    return population.BestProgram;
                }
                else
                {
                    return null;
                }
            }
        }

        /// <summary>
        /// Delegate for the event fired every time a new best solution is found, or every 10 turns.
        /// </summary>
        /// <param name="sender">Object that generated the event.</param>
        /// <param name="e">Event arguments.</param>
        public delegate void IterationCompletedEventHandler(Object sender, IterationCompleteEventArgs e);

        /// <summary>
        /// Event raised every time a new best solution is found, or every 10 turns.
        /// </summary>
        public event IterationCompletedEventHandler IterationCompleted;

        /// <summary>
        /// Default constructor.
        /// </summary>
        static Evolver()
        {
        }

        /// <summary>
        /// Returns a random side, either left or right.
        /// </summary>
        /// <returns>Left or Right</returns>
        private ChildSide RandomSide()
        {
            if (Options.InterpreterRandom(2) == 0)
            {
                return ChildSide.Left;
            }
            else
            {
                return ChildSide.Right;
            }
        }

        /// <summary>
        /// Picks a random node in the program tree to perform crossover or mutation on.
        /// </summary>
        /// <param name="program">The program to find node in.</param>
        /// <param name="minimumSelectionOdds">The odds that at any level that a given node will be selected. Default odds are log2(program size).</param>
        /// <returns>The node to perform crossover or mutation on. Also returns that nodes parent.</returns>
        private InformationForNodeToActOn SelectNodeToActOn(Program program, int minimumSelectionOdds)
        {
            // odds that we will perform crossover at a given level. Log base 2 since this is a binary tree
            int selectionOdds = (int)Math.Ceiling(Math.Log(program.ProgramSize, 2));
            if (selectionOdds < minimumSelectionOdds)
            {
                selectionOdds = minimumSelectionOdds;
            }

            // By default we will act on the parent node.
            InformationForNodeToActOn nodeInfo = new InformationForNodeToActOn();
            nodeInfo.ParentNode = program.ProgramTree;
            nodeInfo.ChildNode = program.ProgramTree;

            // randomly choose child nodes until we hit the end of the tree or until a random point based on the selection odds
            while ((Options.InterpreterRandom(selectionOdds) != 0) && (!Tokens.IsLeafNode(nodeInfo.ChildNode.Instruction)))
            {
                nodeInfo.ParentNode = nodeInfo.ChildNode;

                nodeInfo.Side = RandomSide();
                switch (nodeInfo.Side)
                {
                    case ChildSide.Left:
                        nodeInfo.ChildNode = nodeInfo.ChildNode.Left;
                        break;
                    case ChildSide.Right:
                        nodeInfo.ChildNode = nodeInfo.ChildNode.Right;
                        break;
                }
            }

            return nodeInfo;
        }

        /// <summary>
        /// Swap the node at the crossover point in one of the new programs.
        /// This needs to be called twice to perform the crossover on both new clild programs.
        /// </summary>
        /// <param name="program1">The program that is being updated.</param>
        /// <param name="program1CrossoverPoint">Information on the crossover node in the program being updated.</param>
        /// <param name="program2CrossoverPoint">Information on the crossover node from the program that the substituted nodes are coming from.</param>
        private void PerformCrossover(Program program1, InformationForNodeToActOn program1CrossoverPoint, InformationForNodeToActOn program2CrossoverPoint)
        {
            switch (program1CrossoverPoint.Side)
            {
                case ChildSide.None:
                    program1.ProgramTree = program2CrossoverPoint.ChildNode;
                    break;
                case ChildSide.Left:
                    program1CrossoverPoint.ParentNode.Left = program2CrossoverPoint.ChildNode;
                    break;
                case ChildSide.Right:
                    program1CrossoverPoint.ParentNode.Right = program2CrossoverPoint.ChildNode;
                    break;
            }
        }

        /// <summary>
        /// Performs the crossover operation. The 2 child programs are set as copies of the parent programs.
        /// Then random crossover points are selected in the child programs.
        /// Finally, the program nodes in the binary trees are swapped in the children at the crossover point.
        /// </summary>
        /// <param name="parent1">First program parent.</param>
        /// <param name="parent2">Second program parent.</param>
        /// <param name="child1">First program child.</param>
        /// <param name="child2">Second program child.</param>
        private void Crossover(Program parent1, Program parent2, Program child1, Program child2)
        {
            // The program tree in the children is updated, but the reference to the Program object itself does not change.
            child1.CopyProgram(parent1);
            child2.CopyProgram(parent2);

            InformationForNodeToActOn child1CrossoverPoint = SelectNodeToActOn(child1, 3);
            InformationForNodeToActOn child2CrossoverPoint = SelectNodeToActOn(child2, 3);

            PerformCrossover(child1, child1CrossoverPoint, child2CrossoverPoint);
            PerformCrossover(child2, child2CrossoverPoint, child1CrossoverPoint);

            child1.CalculateSizeOfProgram();
            child2.CalculateSizeOfProgram();
        }

        /// <summary>
        /// Based on percent odds in GeneticProgramOptions, causes mutation to be performed on a program.
        /// One nodes instruction will be replaced with a random instruction of the same type (leaf or non-leaf).
        /// If the mutation is greedy, then the mutation is rejected if it did not improve the program.
        /// </summary>
        /// <param name="program">The program to mutate.</param>
        /// <param name="options">Genetic Program options for the interpreter.</param>
        public void Mutate(Program program, GeneticProgramOptions options)
        {
            // Determine if we should do mutation.
            if (Options.InterpreterRandom(100) > options.PercentMutation)
            {
                return; // don't do mutation this time;
            }

            bool greedyMutation = (Options.InterpreterRandom(100) < options.PercentGreedyMutation);

            InformationForNodeToActOn parentMutationPoint = SelectNodeToActOn(program, 2);

            long oldFitness = program.Fitness;
            int oldFoodLeft = program.FoodLeft;
            int oldFoodNotFound = program.FoodNotFound;
            int oldWaterRemaining = program.WaterRemaining;
            int oldNumberTurns = program.NumberOfTurns;

            Tokens.Instruction oldNodeInstruction = parentMutationPoint.ParentNode.Instruction;

            if (Tokens.IsLeafNode(parentMutationPoint.ParentNode.Instruction))
            {
                if (parentMutationPoint.Side == ChildSide.None)
                {
                    program.ProgramTree.Instruction = Tokens.GetRandomLeafNode(options);
                }
                else
                {
                    parentMutationPoint.ParentNode.Instruction = Tokens.GetRandomLeafNode(options);
                }
            }
            else
            {
                if (parentMutationPoint.Side == ChildSide.None)
                {
                    program.ProgramTree.Instruction = Tokens.GetRandomNonLeafNode(options);
                }
                else
                {
                    parentMutationPoint.ParentNode.Instruction = Tokens.GetRandomNonLeafNode(options);
                }
            }

            if (population.TestNewProgram(program, Options))
            {
                onIterationComplete(population);  // This program is the best in the population, so raise an event to the GUI.
            }

            if ((greedyMutation) && (oldFitness < program.Fitness))
            {
                // if the user wanted greedy mutation (only accept the mutation if it improved the program)
                // and this mutation did not improve the fitness, then reject the mutation.
                if (parentMutationPoint.Side == ChildSide.None)
                {
                    program.ProgramTree.Instruction = oldNodeInstruction;
                }
                else
                {
                    parentMutationPoint.ParentNode.Instruction = oldNodeInstruction;
                }
                // also reset the fitness back to its old value.
                program.Fitness = oldFitness;
                program.FoodLeft = oldFoodLeft;
                program.FoodNotFound = oldFoodNotFound;
                program.WaterRemaining = oldWaterRemaining;
                program.NumberOfTurns = oldNumberTurns;
            }
        }

        /// <summary>
        /// Raise an event back to the GUI.
        /// </summary>
        /// <param name="programs">The population used by the Genetic Program.</param>
        private void onIterationComplete(Population programs)
        {
            if (IterationCompleted != null)
            {
                IterationCompleteEventArgs e = new IterationCompleteEventArgs(programs.BestProgram, programs.AverageFitness, programs.AverageProgramSize, programs.Count, iteration);
                IterationCompleted(this, e);
            }
        }

        /// <summary>
        /// This is the primary function in performing the Genetic Program.
        /// It is expected that the GeneticProgramOptions on this object are set before starting the evolver.
        /// </summary>
        /// <param name="stateInfo">Standard parameter for a function started on a thread pool thread using QueueUserWorkItem.</param>
        private void StartRunningEvolver(Object stateInfo)
        {
            Running = true;
            
            while (Running)
            {
                if ((iteration % 10) == 0)
                {
                    onIterationComplete(population);
                }

                iteration++;

                // Select N random programs from the population, and sort them by fitness.
                List<Program> tournament = population.SelectTournament(Options);

                // The best 2 programs in the tournament are the 2 parents of the new programs,
                // and the worst 2 programs in the tournament are replaced with the new child programs.
                Crossover(tournament[0], tournament[1], tournament[tournament.Count - 2], tournament[tournament.Count - 1]);

                // test the first child. Raise an event if this program is the best in the population.
                if (population.TestNewProgram(tournament[tournament.Count - 2], Options))
                {
                    onIterationComplete(population);
                }

                // test the second child. Raise an event if this program is the best in the population.
                if (population.TestNewProgram(tournament[tournament.Count - 1], Options))
                {
                    onIterationComplete(population);
                }

                Mutate(tournament[tournament.Count - 2], Options);
                Mutate(tournament[tournament.Count - 1], Options);
            }

            onIterationComplete(population);  // after the user stop the genetic program, fire one last update to the GUI.
        }

        /// <summary>
        /// This function resumes the Genetic Program.
        /// It is expected that EvolveSolution has already been run once on the Evolver, and this program is resuming the Genetic Program.
        /// If the user updates the GeneticProgramOptions object, then those new options will be used.
        /// </summary>
        /// <param name="stateInfo">Standard parameter for a function started on a thread pool thread using QueueUserWorkItem.</param>
        public void ResumeSolution(Object stateInfo)
        {
            int changeInPopulationSize = options.PopulationSize - population.Count;

            Running = true;

            if (changeInPopulationSize > 0) // The user has increased the population size.
            {
                for (int i = 0; i < changeInPopulationSize; i++)
                {
                    if (!Running)   // Has the user hit the stop button
                    {
                        onIterationComplete(population);
                        return;
                    }

                    Program program = new Program(options);
                    population.TestNewProgram(program, options);
                    population.Add(program);

                    if (population.Count % 10 == 0)    // After every 10 programs are created, raise an event to the GUI.
                    {
                        onIterationComplete(population);
                    }
                }
            }
            else if (changeInPopulationSize < 0) // The user has increased the population size.
            {
                population.RemoveRange(population.Count + changeInPopulationSize, -changeInPopulationSize);

                population.FindBestProgram();
            }

            StartRunningEvolver(stateInfo);
        }

        /// <summary>
        /// This function starts the Genetic Program fresh.
        /// It is expected that the GeneticProgramOptions on this object are set before starting the evolver.
        /// </summary>
        /// <param name="stateInfo">Standard parameter for a function started on a thread pool thread using QueueUserWorkItem.</param>
        public void EvolveSolution(Object stateInfo)
        {
            population = new Population();

            Running = true;

            for (int i = 0; i < options.PopulationSize; i++)
            {
                if (!Running) // Has the user hit the stop button
                {
                    onIterationComplete(population);
                    return;
                }

                if (i % 10 == 0)   // After every 10 programs are created, raise an event to the GUI.
                {
                    onIterationComplete(population);
                }

                Program program = new Program(options);
                population.TestNewProgram(program, options);
                population.Add(program);
            }
        
            iteration = 0;

            StartRunningEvolver(stateInfo);
        }
    }
}
