//////////////////////////////////////////////////////////////////////////////////////////////////
// File Name: Interpreter.cs
//      Date: 06/01/2006
// Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
// Permission to use, copy, modify, and distribute this Program and its documentation,
//  if any, for any purpose and without fee is hereby granted, provided that:
//   (i) you not charge any fee for the Program, and the Program not be incorporated
//       by you in any software or code for which compensation is expected or received;
//   (ii) the copyright notice listed above appears in all copies;
//   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
//   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
//          pertaining to distribution of the Program without specific, written prior permission. 
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;

namespace AntGP
{
	/// <summary>
	/// This is the class that interprets the ant program.
	/// </summary>
	public class Interpreter
	{
        /// <summary>
        /// Delegate used when each time the interpreter finishes a turn.
        /// </summary>
        /// <param name="sender">Object that generated this event.</param>
        /// <param name="e">Event Arguments.</param>
		public delegate void TurnCompletedEventHandler(Object sender, TurnCompleteEventArgs e);

        /// <summary>
        /// Event raised when the interpreter finishes a turn.
        /// </summary>
		public event TurnCompletedEventHandler TurnCompleted;

        /// <summary>
        /// Default constructor.
        /// </summary>
		static Interpreter()
		{
		}
        
        /// <summary>
        /// The colony of ants that are executing the ant program.
        /// </summary>
        Colony ants;

        /// <summary>
        /// The map that the ants are searching for food on.
        /// </summary>
        Map map;

        /// <summary>
        /// Private for the Genetic Program options used by the interpreter.
        /// </summary>
        private GeneticProgramOptions options;
        /// <summary>
        /// The Genetic Program options used by the interpreter.
        /// </summary>
        public GeneticProgramOptions Options
        {
            set
            {
                options = value;
            }
            get
            {
                return options;
            }
        }

        /// <summary>
        /// Private for the program being run by the ants.
        /// </summary>
        private Program program;
        /// <summary>
        /// The program being run by tht ants.
        /// </summary>
        public Program Program
        {
            set
            {
                program = value;
            }
            get
            {
                return program;
            }
        }

        /// <summary>
        /// Private flag for determining if the interpreter running.
        /// </summary>
		private bool running = false;
        /// <summary>
        /// Should the interpreter continue running.
        /// The GUI can request the interpreter to stop by setting this flag to false.
        /// </summary>
		public bool Running
		{
			set 
			{
				running = value;
			}
			get 
			{
				return running;
			}
		}

		private void onTurnComplete(Map map, int numberOfNodes, int turn, int averageAntMovement, bool complete)
		{
			if ( TurnCompleted != null )
			{
                TurnCompleteEventArgs e = new TurnCompleteEventArgs(map, ants, numberOfNodes, turn, averageAntMovement, complete);
				TurnCompleted( this, e );
			}
		}

        public void ExecuteProgram(Object stateInfo)
		{
            map = new Map(Options.Map);  // must make a copy of the map, becuase we will be moving the food.
            ants = new Colony(Options.NumberOfAnts, map);

            int numberOfNodes = program.ProgramSize;

			int moves = 0;
			int turns = 0;
            Running = true;

			onTurnComplete(map, numberOfNodes, 0, 0, false);

			while ((Running) && (moves < Options.MaximumMovementForAllAntsPerTest) && (!map.IsAllFoodCollected())
                && (turns < Options.NumberOfTurnsPerAnt) && (ants.TotalMovement < Options.MaximumMovementForAllAntsPerTest))
			{
				if ( Options.PheromoneDissipationSpeed > 0 )
				{
					map.SpreadPheromones(Options.PheromoneDissipationSpeed);
				}

				map.ResetFoodForNewTurn();

				foreach( Ant ant in ants )
				{
					if ( ant.IsAlive )
					{
                        if ((ant.FoodCarried != null) && (ant.FoodCarried.MovedThisTurn))
                        {
                            // If another ant carrying this same food has already moved this turn,
                            // then just move to the same location as that ant.
                            ant.MoveWithFoodCarryingAntLeader();
                        }
                        else
                        {
                            ant.Options = Options;
                            executeProgramRecursive(Program.ProgramTree, ant);
                        }
					}
				}

                turns++;
                onTurnComplete(map, numberOfNodes, turns, (int)(ants.TotalMovement / ants.Count), false);
			}

            Running = false;

            program.FoodLeft = map.FoodNotCollected();
            program.FoodNotFound = map.FoodNotLocated();
            program.WaterRemaining = map.WaterRemaining();

            int averageAntMovement = Options.NumberOfTurnsPerAnt;
            if (map.IsAllFoodCollected())
            {
                // Only for solutions that work do we start to reward finishing early.
                // Otherwise, when comparing two solutions that find no food, we reward the one where the ants don't leave the nest.
                averageAntMovement = (int)(ants.TotalMovement / ants.Count);
            }

            program.NumberOfTurns = averageAntMovement;

            onTurnComplete(map, numberOfNodes, turns, averageAntMovement, true);
		}
        
        /// <summary>
        /// Recursively executes the nodes in the ant program.
        /// </summary>
        /// <param name="node">Current node being interpreted.</param>
        /// <param name="ant">The ant that is running the program.</param>
        private void executeProgramRecursive(ProgramNode node, Ant ant)
		{
            if (!ant.IsAlive)   // If the ant died in a previous instruction, stop running the program.
            {
                return;
            }
			if (Tokens.IsLeafNode(node.Instruction))
			{
				processInstruction(node.Instruction, ant);
			}
			else
			{
                if (Tokens.IsIfStatement(node.Instruction))
                {
                    if (evaluateCondition(node.Instruction, ant)) // For if statements, evaluate the condition and then determine what child node to execute.
                    {     
			            if (node.Left != null)
			            {
				            executeProgramRecursive(node.Left, ant);
			            }
                    }
                    else
                    {
				        if (node.Right != null)
				        {
					        executeProgramRecursive(node.Right, ant);
				        }
                    }
                }
                else if ( node.Instruction == Tokens.Instruction.TwoStatements )
                {   // For TwoStatements, both sides of the binary tree are executed.
                    if (node.Left != null)
                    {
                        executeProgramRecursive(node.Left, ant);
                    }
                    if (node.Right != null)
                    {
                        executeProgramRecursive(node.Right, ant);
                    }
                }
			}
		}

        /// <summary>
        /// Execute an if statement node in the ant program and determine if the condition is true.
        /// </summary>
        /// <param name="instruction">The instruction to execute</param>
        /// <param name="ant">The ant that is executing the instruction.</param>
        /// <returns>Is the condition true.</returns>
        private bool evaluateCondition(Tokens.Instruction instruction, Ant ant)
		{
            switch( instruction )
            {
                case Tokens.Instruction.IfFoodAhead:
                    return ant.IsItemAhead(MapItem.Item.AnyFood);
                case Tokens.Instruction.IfFoodLeft:
                    return ant.IsItemAdjacent(MapItem.Item.AnyFood, Compass.LeftDirection(ant.Direction));
                case Tokens.Instruction.IfFoodRight:
                    return ant.IsItemAdjacent(MapItem.Item.AnyFood, Compass.RightDirection(ant.Direction));
                case Tokens.Instruction.IfFoodAdjacent:
                    return ant.IsItemAdjacent(MapItem.Item.AnyFood) != Compass.CompassDirection.None;
                case Tokens.Instruction.IfFoodHere:
                    return ant.IsItemHere(MapItem.Item.AnyFood);
                case Tokens.Instruction.IfPheromoneHere:
                    return ant.IsPheromoneHere();
                case Tokens.Instruction.IfLeafHere:
                    return ant.IsItemHere(MapItem.Item.Leaf);
                case Tokens.Instruction.IfWaterAhead:
                    return ant.IsItemAhead(MapItem.Item.Water);
                case Tokens.Instruction.IfObstacleAhead:
                    return ant.IsItemAhead(MapItem.Item.Obstacle);
                case Tokens.Instruction.IfAtNest:
                    return ant.IsItemHere(MapItem.Item.Nest);
                case Tokens.Instruction.IfMoveToAdjacentFood:
                    return ant.MoveToAdjacentItem(MapItem.Item.AnyFood);
                case Tokens.Instruction.IfMoveToAdjacentLeaf:
                    if (!ant.MoveToAdjacentItem(MapItem.Item.Tree))
                    {
                        return ant.MoveToAdjacentItem(MapItem.Item.Leaf);
                    }
                    return true;
                case Tokens.Instruction.IfMoveToStrongestPheromone:
                    return ant.MoveToStongestPheromone();
                case Tokens.Instruction.IfMoveToPheromoneAwayFromNest:
                    return ant.MoveToStrongestPheromoneAwayFromNest();
                case Tokens.Instruction.IfMoveToDeadAnt:
                    return ant.MoveToAdjacentItem(MapItem.Item.DeadAnt);
                case Tokens.Instruction.IfMoveToNest:
                    return ant.MoveToNest();
                case Tokens.Instruction.IfMoveToNestThroughWater:
                    if (!ant.MoveToNest())  // if move fails, we assume this function has side effect of pointing us in right direction
                    {
                        if (!ant.MoveIntoWater())
                        {
                            return false;
                        }
                    }
                    return true;
                case Tokens.Instruction.IfCarryingFood:
                    return ant.FoodCarried != null;
                case Tokens.Instruction.IfCarryingLeaf:
                    return ant.CarryingLeaf;
                case Tokens.Instruction.IfAnotherAntHere:
                    foreach (Ant otherAnt in ants)
                    {
                        if ((otherAnt.Location == ant.Location) && (otherAnt != ant))
                        {
                            return true;
                        }
                    }
                    return false;
                case Tokens.Instruction.IfCantLiftFood:
                    return ant.FoodTooHeavy();
                case Tokens.Instruction.IfTiredOfWaiting:
                    if (ant.FoodTooHeavy())
                    {
                        if (options.InterpreterRandom(100) > Options.WaitForHelpThisTurnPercent)
                        {
                            return true;
                        }
                    }
                    return false;
            }
            return false;
		}

        /// <summary>
        /// Executes a leaf statement for the given ant.
        /// </summary>
        /// <param name="instruction">The statement to execute.</param>
        /// <param name="ant">The ant that is executing the instruction.</param>
		private void processInstruction(Tokens.Instruction instruction, Ant ant)
		{
            switch( instruction )
            {
                case Tokens.Instruction.TurnLeft:
                    ant.TurnLeft();
                    break;
                case Tokens.Instruction.TurnRight:
                    ant.TurnRight();
                    break;
                case Tokens.Instruction.MoveForward:
                    ant.MoveForward();
                    break;
                case Tokens.Instruction.MoveRandom:
                    ant.Direction = Compass.RandomDirection(options);
                    ant.MoveForward();
                    break;
                case Tokens.Instruction.MoveQuasiRandom:
            		/* 50% straight, 20% left or right, 10% behind */
                    int randDir = options.InterpreterRandom(10);
                    if ( randDir < 1)
                    {
                        ant.TurnRight();
                        ant.TurnRight();
                    }
                    else if (randDir < 3)
                    {
                        ant.TurnRight();
                    }
                    else if (randDir < 5)
                    {
                        ant.TurnLeft();
                    }
                    ant.MoveForward();
                    break;
                case Tokens.Instruction.MoveToNest:
                    ant.MoveToNest();
                    break;
                case Tokens.Instruction.MoveToNestThroughWater:
                    if (!ant.MoveToNest())  // if move fails, we assume this function has side effect of pointing us in right direction
                    {
                        ant.MoveIntoWater();
                    }
                    break;
                case Tokens.Instruction.MoveIntoWater:
                    ant.MoveIntoWater();
                    break;
                case Tokens.Instruction.PickUpFood:
                    ant.PickUpFood();
                    break;
                case Tokens.Instruction.DropFood:
                    ant.DropFood();
                    break;
                case Tokens.Instruction.PickUpLeaf:
                    ant.PickUpLeaf();
                    break;
                case Tokens.Instruction.DropLeaf:
                    ant.DropLeaf();
                    break;
                case Tokens.Instruction.DropLeafInWater:
                    ant.DropLeafInWater();
                    break;
                case Tokens.Instruction.ReleasePheromone:
                    ant.ReleasePheromone(Options.PheromoneStrength);            
                    break;
                case Tokens.Instruction.EscapePheromone:
                    ant.EscapePheromone(options);
                    break;
            }
		}
	}
}